#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <libc.h>
#include <string.h>

#include "String.h"

extern "C" {
	int strncasecmp(const char*, const char*, int);
}

RJS_String NullRJS_String;	// used for NULL substrings

RJS_StringRange RJS_String::hunksize=16; // number to round up new operator requests
RJS_String::CaseCompare RJS_String::use_case=RJS_String::UseCase; // use case when comparing
char   RJS_String::temp_icase=0;
char   RJS_String::temp_case=0;
RJS_String::UpdateFunc RJS_String::uf=NULL;

void safe_memcpy(char *dest, char *src, int count) 
{

  if (dest > src && dest < (src+count)) { // overlap, copy in reverse
	while (count--) *(dest+count) = *(src+count);
  } else { // ok, copy normally
	memcpy(dest,src,count);
  }
}


// this function calculates the size of a new request
// should be passed the size of the string to be allocated,
// not including the NULL

// Utility functions 

int hunk_size(int n) { 
int temp;
  n++; // add one for null
  temp = (n/RJS_String::get_hunksize())*RJS_String::get_hunksize()+
	RJS_String::get_hunksize(); 
  if (temp > RJS_StringMax) {
	cerr << "RJS_String Exception => maximum length exceeded" << endl;
	exit(1);
  }
  return temp;
}

// handle execeptions

void RJS_String::new_error(char *message) 
{
  cerr << "RJS_String: new failed ==> " << message << endl ;
  exit(1);
}

void RJS_String::exception(char *message)  const
{
  cerr << "RJS_String Exception => " << message << endl << endl;
  dump();
  exit(1);
}


// dump out string contents

void RJS_String::dump() const
{
  cerr << "RJS_String length: " << sd.len << endl;
  cerr << "RJS_String size  : " << sd.size << endl;
  cerr << "RJS_String data  : " << endl;
  for (int i=0; i<sd.len; i++) 
	cerr << setw(5) << i << " :   " << sd.data[i] << " (" << int(sd.data[i]) << ")" << endl;
}

// allocate a string
RJS_StringData RJS_String::alloc(const char *data, int len, int size)
{
RJS_StringData temp;
  
  temp.size = (size== -1) ? hunk_size(len) : hunk_size(size);
  temp.len=len;
  temp.data = new char[temp.size];
  if (len) memcpy(temp.data,data,len);  
  temp.data[len]='\0';
  return temp;
}

void RJS_String::resize(RJS_StringData &sd, int new_size) 
{
  if (new_size >= sd.size) {
	RJS_StringData temp=RJS_String::alloc(sd.data,sd.len,new_size);
	delete sd.data;
	sd=temp;
  }
}


void RJS_String::assign(const char *data, RJS_StringRange len)
{
  if (this==&NullRJS_String) return;

  if (len >= sd.size) {
	delete sd.data;
	sd=RJS_String::alloc(data,len);
  } else {
	memcpy(sd.data,data,len);
	sd.data[len]='\0';
	sd.len=len;
  }
  update();
}

// This function has the task of assign a substring a new value.
//
// string:  xxxxxx yyyyyyyy zzzzzzz
//
// len_before = len(x's)     length of string before substring
// len = len(y's)            length of substring 
// len_after = len(z's)      length of string after substring
//

void RJS_SubString::assign(const char *ndata,RJS_StringRange nlen)
{

if (str==&NullRJS_String) return;

int olen=length();

if (olen==0 && nlen==0) return; // nothing to do!

RJS_StringData &str_sd = (str->sd);
RJS_SubString *stmp=sstr; // list of RJS_SubStrings

int len_before = cptr()-str->cptr();		// len before substring
int len_after  = str_sd.len - (len_before+olen);    // len after substring
int new_len    = len_before + nlen + len_after; // new length of string
int diff = nlen-olen;
/*
cerr << " len_before = " << len_before << endl;
cerr << " len_after  = " << len_after << endl;
cerr << " nlen       = " << nlen << endl;
cerr << " olen       = " << olen << endl;
cerr << " new_len    = " << new_len << endl;
*/
// if we need to resize string, just make a new copy
if (new_len >= str_sd.size) goto new_copy; 

// check of we are just appending
if (olen==0 && len_before==str_sd.len) { 
	memcpy(sd.data,ndata,nlen);
	str_sd.len=new_len;
	str_sd.data[str_sd.len]='\0';
	sd.size=sd.len=nlen; // update RJS_SubString
	while(stmp) {	// update parent RJS_SubString's
		stmp->sd.size = ( stmp->sd.len += diff);
		stmp=stmp->sstr;
	}
	str->update();
   	return;
}

if (nlen==0) { // removing data
  if (len_after) memcpy(str_sd.data+len_before,str_sd.data+len_before+olen,len_after);
  str_sd.len = len_before+len_after;
  str_sd.data[str_sd.len]='\0';
  sd.size=sd.len=0; // update RJS_SubString
  while(stmp) {	// update parent RJS_SubString's
		stmp->sd.size = ( stmp->sd.len += diff);
		stmp=stmp->sstr;
  }
  str->update();
  return;
}

// check if new value overlaps any old data, of so don't bother, just copy it!
if (ndata >= str_sd.data && ndata < str_sd.data+str_sd.len) goto new_copy;

if (nlen==olen) {	// same size
	memcpy(str_sd.data+len_before,ndata,nlen);
} else {
	safe_memcpy(sd.data+nlen,sd.data+olen,len_after);
	memcpy(sd.data,ndata,nlen);
	str_sd.len=new_len;
	str_sd.data[str_sd.len]='\0';
	sd.len=sd.size=nlen;
 	while(stmp) {	// update parent RJS_SubString's
		stmp->sd.size = ( stmp->sd.len += diff);
		stmp=stmp->sstr;
	}
}

  str->update();
  return;

new_copy:

RJS_StringData temp;
temp.size = hunk_size(new_len);
temp.len  = new_len;
temp.data = new char[temp.size];

// make new string, copy before, substr, after

if (len_before) 
	memcpy(temp.data,str_sd.data,len_before);
if (nlen) 
	memcpy(temp.data+len_before,ndata,nlen);
if (len_after) 
	memcpy(temp.data+len_before+nlen,str_sd.data+len_before+olen,len_after);

  temp.data[new_len] = '\0';
  char *old_cptr=str_sd.data; // save if needed for substrings
  delete str_sd.data; // delete old string
  str_sd = temp;

  sd.data = str_sd.data+len_before;
  sd.size = sd.len = nlen;

  while(stmp) {
	stmp->sd.data = str_sd.data + (stmp->sd.data-old_cptr);
	stmp->sd.size = ( stmp->sd.len += diff );
	stmp=stmp->sstr;
  }
  str->update();
}

int RJS_String::length(const char *s) 
{
const char *p=s;

	while (*p) p++;
	return int(p-s);
}



int RJS_String::search(const char *search_str,int sslen) const
{
const char *source=sd.data;
int slen=sd.len;

 if ((use_case==UseCase && !temp_icase) || 
  (use_case==IgnoreCase && !temp_icase && temp_case)) {
   if (temp_case) temp_case=0;
   if (temp_icase) temp_icase=0;
   if (sslen==0 && sd.len!=0) return -1; 
   if (sslen==1) {
	const char *loc=source; char search_ch = *search_str;
        while (slen--)
             if ( *loc == search_ch) return loc-source;
             else loc++;
        return -1;
   }

  const char *stop_index=source+slen-sslen+1; // give up if not found yet
  const char *curr_index=source;            // starting point

  if (stop_index> curr_index) while(curr_index != stop_index) {
     register const char *sp=curr_index;
     register const char *ssp=search_str;
     register const char *ss_end=ssp+sslen;
     while ((ssp != ss_end) && (*sp++ == *ssp++)) ;   // spin over string
     if (*(sp-1)== *(ssp-1)) return curr_index-source; // check for equal
     else ++curr_index;
  }
 } else {
   if (temp_case) temp_case=0;
   if (temp_icase) temp_icase=0;
   if (sslen==0 && sd.len!=0) return -1; 
   if (sslen==1) {
	const char *loc=source; char search_ch = ToUpper(*search_str);
        while (slen--)
             if ( ToUpper(*loc) == search_ch) return loc-source;
             else loc++;
        return -1;
   }
  const char *stop_index=source+slen-sslen+1; // give up if not found yet
  const char *curr_index=source;            // starting point
  if (stop_index> curr_index) while(curr_index != stop_index) {
     register const char *sp=curr_index;
     register const char *ssp=search_str;
     register const char *ss_end=ssp+sslen;
     while ((ssp != ss_end) && (ToUpper(*sp++) == ToUpper(*ssp++))) ;   // spin over string
     if (ToUpper(*(sp-1))== ToUpper(*(ssp-1))) return curr_index-source; // check for equal
     else ++curr_index;
  }
 }
 return -1;	// not found
}

int RJS_String::compare(char ch) const
{
int diff;

  if ((use_case==UseCase && !temp_icase) || 
      (use_case==IgnoreCase && !temp_icase && temp_case))
		diff=sd.data[0]-ch;
  else
		diff=ToUpper(sd.data[0])-ToUpper(ch);

  if (temp_case) temp_case=0;
  if (temp_icase) temp_icase=0;

  if (diff !=0 || length()==1) return diff;
  else return (length()>1) ? 1 : -1;
}


int RJS_String::compare(const char *s2) const
{
int s2_len=RJS_String::length(s2);
int diff;


  if ((use_case==UseCase && !temp_icase) || 
      (use_case==IgnoreCase && !temp_icase && temp_case))
	  diff=strncmp(cptr(),s2,length());  
  else
	  diff=strncasecmp(cptr(),s2,length());  

  if (temp_case) temp_case=0;
  if (temp_icase) temp_icase=0;

  if (diff !=0 || length()==s2_len) return diff;
  else return (length()>s2_len) ? 1 : -1;
}

int RJS_String::compare(const RJS_String &s2) const
{
int diff;

  if ((use_case==UseCase && !temp_icase) || 
      (use_case==IgnoreCase && !temp_icase && temp_case))
	  diff=strncmp(cptr(),s2.cptr(),length());
  else
	  diff=strncasecmp(cptr(),s2.cptr(),length());

  if (temp_case) temp_case=0;
  if (temp_icase) temp_icase=0;

  if (diff !=0 || length()==s2.length()) return diff;
  else return (length()>s2.length()) ? 1 : -1;
}


int RJS_String::compare(const char *s1,int s1_len,const char *s2,int s2_len)
{
int  diff;

  if ((use_case==UseCase && !temp_icase) || 
      (use_case==IgnoreCase && !temp_icase && temp_case))
	diff=strncmp(s1,s2,s1_len);
  else
	diff=strncasecmp(s1,s2,s1_len);

  if (temp_case) temp_case=0;
  if (temp_icase) temp_icase=0;
  
  if (diff !=0 || s1_len==s2_len) return diff;
  else return (s1_len>s2_len) ? 1 : -1;
}


void RJS_String::VERIFY() const
{
if (sd.len >= sd.size) exception("VERIFY :  length >= size");
if (sd.data[sd.len] != '\0') exception("VERIFY :  missing null");
if (sd.data==NULL) exception("VERIFY : data == NULL");
}

ushort RJS_String::set_length(ushort n)
{
  if (n < sd.size) trunc(n);
  return sd.len;
}

