#ifndef RJS_STRING_CLASS_H
#define RJS_STRING_CLASS_H

// 
// Classname: RJS_String 
//
// Author: Roland Schemers 
//
// Date : June 6th, 1990
//
//

#include <iostream.h>

const RJS_StringMax=65535; //including NULL

typedef unsigned short RJS_StringRange;

typedef struct {
	RJS_StringRange	len;		// length of string (not including null)
	RJS_StringRange	size;		// allocated size of string
	char 		*data;		// pointer to data
} RJS_StringData;

class RJS_StringSearch;
class RJS_String;

class RJS_String {
	friend ostream &operator<<(ostream &os,const RJS_String &s);
	friend istream &operator>>(istream &is, RJS_String &s);
	friend int getline(istream &is,RJS_String &s,char delim='\n',int dropit=1);
	friend class RJS_SubString;
public:

enum Side { Left=1,Right=2,Both=3 };

enum CaseCompare { IgnoreCase=0,UseCase=1};

typedef void (*UpdateFunc)(RJS_String *);

// Constructors
	RJS_String();		
	RJS_String(int n);
	RJS_String(char ch);
	RJS_String(const char *s);
	RJS_String(const char *s,int n);
	RJS_String(const RJS_String &s);
private:  // for RJS_SubString class use only:
	RJS_String(char *s,RJS_StringRange len,RJS_StringRange size);
public:

// Destructor

	~RJS_String() { delete sd.data; }

// relational

	int operator <  (char ch)          const { return compare(ch) < 0; }
	int operator <  (const char *c)    const { return compare(c)  < 0; }
	int operator <  (const RJS_String &s2) const { return compare(s2) < 0; }

	int operator <= (char ch)          const { return compare(ch) <= 0; }
	int operator <= (const char *c)    const { return compare(c)  <= 0; }
	int operator <= (const RJS_String &s2) const { return compare(s2) <= 0; }

	int operator == (char ch)          const { return compare(ch) == 0; }
	int operator == (const char *c)    const { return compare(c)  == 0; }
	int operator == (const RJS_String &s2) const { return compare(s2) == 0; }

	int operator != (char ch)          const { return compare(ch) != 0; }
	int operator != (const char *c)    const { return compare(c)  != 0; }
	int operator != (const RJS_String &s2) const { return compare(s2) != 0; }

	int operator > (char ch)           const { return compare(ch) > 0; }
	int operator > (const char *c)     const { return compare(c)  > 0; }
	int operator > (const RJS_String &s2)  const { return compare(s2) > 0; }

	int operator >= (char ch)          const { return compare(ch) >= 0; }
	int operator >= (const char *c)    const { return compare(c)  >= 0; }
	int operator >= (const RJS_String &s2) const { return compare(s2) >= 0; }

// = operator 

	RJS_String &operator = (char ch);
	RJS_String &operator = (const char *s);
	RJS_String &operator = (const RJS_String &s);

	RJS_String &operator += (char ch);
	RJS_String &operator += (const char *s);
	RJS_String &operator += (const RJS_String &s);

	RJS_String &operator -= (int i) ;
	RJS_String &operator -= (char ch);
	RJS_String &operator -= (const char *s);
	RJS_String &operator -= (const RJS_String &s);
	RJS_String &operator -= (const RJS_StringSearch &ss);

	RJS_String &operator *= (int n);

	RJS_String &operator /= (char ch);
	RJS_String &operator /= (const char *s);
	RJS_String &operator /= (const RJS_String &s);
	RJS_String &operator /= (const RJS_StringSearch &ss);

	RJS_String operator + (char ch) const;
	RJS_String operator + (const char *s) const;
	RJS_String operator + (const RJS_String &s) const;
	friend RJS_String operator + (char ch, const RJS_String &s);
	friend RJS_String operator + (const char *s, const RJS_String &s);

	RJS_String operator - (int n) const;
	RJS_String operator - (char ch) const;
	RJS_String operator - (const char *s) const;
	RJS_String operator - (const RJS_String &s) const;
	RJS_String operator - (const RJS_StringSearch &ss) const;

	RJS_String operator / (char ch) const;
	RJS_String operator / (const char *s) const;
	RJS_String operator / (const RJS_String &s) const;
	RJS_String operator / (const RJS_StringSearch &ss) const;

	int index(char ch) const;
	int index(const char *s) const;
	int index(const RJS_String &s) const;
	int index(const RJS_StringSearch &ss) const;

	int contains(char ch) const;
	int contains(const char *s) const;
	int contains(const RJS_String &s) const;
	int contains(const RJS_StringSearch &ss) const;

// RJS_SubString functions

	char operator[](int i) const; // use substr(i,1) instead of (char &[])

	virtual RJS_SubString substr(int p, int n); 
	RJS_SubString substr(int p);

	RJS_SubString left(int n);
	RJS_SubString right(int n);
	RJS_SubString between(int p1,int p2);

	RJS_String &insert(int pos,char c);
	RJS_String &insert(int pos,const char *s);
	RJS_String &insert(int pos,const RJS_String &s);

	RJS_String &prepend(char c);
	RJS_String &prepend(const char *s);
	RJS_String &prepend(const RJS_String &s);

	RJS_String &append(char c);
	RJS_String &append(const char *s);
	RJS_String &append(const RJS_String &s);

	RJS_String &remove(int pos, int n);
	RJS_String &remove(char c);
	RJS_String &remove(const char *);
	RJS_String &remove(const RJS_String &s);
	RJS_String &remove(const RJS_StringSearch &ss);

	RJS_SubString operator () (int pos);
	RJS_SubString operator () (int pos, int n);
	RJS_SubString operator () (char c);
	RJS_SubString operator () (const char *s);
	RJS_SubString operator () (const RJS_String &s);
	RJS_SubString operator () (const RJS_StringSearch &ss);

// before -- up to but not including...
	RJS_SubString before(int pos);
	RJS_SubString before(char c);
	RJS_SubString before(const char *s);
	RJS_SubString before(const RJS_String &s);
	RJS_SubString before(const RJS_StringSearch &ss);

// through  -- up to and including...
	RJS_SubString through(int pos);
	RJS_SubString through(char c);
	RJS_SubString through(const char *s);
	RJS_SubString through(const RJS_String &s);
	RJS_SubString through(const RJS_StringSearch &ss);

// at  -- at this position only...
	RJS_SubString at(int pos,int n=1);
	RJS_SubString at(char c);
	RJS_SubString at(const char *s);
	RJS_SubString at(const RJS_String &s);
	RJS_SubString at(const RJS_StringSearch &ss);

// from  -- from this position to the end...
	RJS_SubString from(int pos);
	RJS_SubString from(char c);
	RJS_SubString from(const char *s);
	RJS_SubString from(const RJS_String &s);
	RJS_SubString from(const RJS_StringSearch &ss);

// after  -- after this position to the end
	RJS_SubString after(int pos);
	RJS_SubString after(char c);
	RJS_SubString after(const char *s);
	RJS_SubString after(const RJS_String &s);
	RJS_SubString after(const RJS_StringSearch &ss);

// except -- return everything in string except
	RJS_String except(int pos,int n);
	RJS_String except(char c);
	RJS_String except(const char *s);
	RJS_String except(const RJS_String &s);
	RJS_String except(const RJS_StringSearch &ss);

// Misc functions 

	RJS_SubString skip(char c);
	RJS_SubString skip(const char *s);
	RJS_SubString skip(const RJS_String &s);
	RJS_SubString skip(const RJS_StringSearch &ss);
	RJS_SubString skip();			// skips white space
	RJS_SubString ws();				// skips white space

	RJS_String &replace(char c1, char c2);
	RJS_String &replace(char c1, const char *s);
	RJS_String &replace(char c1, const RJS_String &s);
	RJS_String &replace(const char *s1, char c1);
	RJS_String &replace(const char *s1, const char *s2);
	RJS_String &replace(const char *s1, const RJS_String &s2);
	RJS_String &replace(const RJS_String &s1, char c1);
	RJS_String &replace(const RJS_String &s1, const char *s2);
	RJS_String &replace(const RJS_String &s1, const RJS_String &s2);
	RJS_String &replace(const RJS_StringSearch &ss, char c1);
	RJS_String &replace(const RJS_StringSearch &ss, const char *s);
	RJS_String &replace(const RJS_StringSearch &ss, const RJS_String &s);
	RJS_String &replace(const RJS_StringSearch &ss, char c1, RJS_String &matched);
	RJS_String &replace(const RJS_StringSearch &ss, const char *s, RJS_String &matched);
	RJS_String &replace(const RJS_StringSearch &ss, const RJS_String &s, RJS_String &matched);

	RJS_String&   pos(int &pos);

	RJS_SubString moveto(char ch);
	RJS_SubString moveto(char ch,int &pos);
	RJS_SubString moveto(const char *s);
	RJS_SubString moveto(const char *s,int &pos);
	RJS_SubString moveto(RJS_String &s);
	RJS_SubString moveto(RJS_String &s,int &pos);
	RJS_SubString moveto(const RJS_StringSearch &ss);
	RJS_SubString moveto(const RJS_StringSearch &ss,int &pos, int &n);
	RJS_SubString moveto(const RJS_StringSearch &ss, RJS_String &matched); // s=matched string
	RJS_SubString moveto(const RJS_StringSearch &ss, RJS_String &matched, int &pos, int &n);

	RJS_SubString find(char ch);
	RJS_SubString find(char ch,int &pos);
	RJS_SubString find(const char *s);
	RJS_SubString find(const char *s,int &pos);
	RJS_SubString find(const RJS_String &s);
	RJS_SubString find(const RJS_String &s,int &pos);
	RJS_SubString find(const RJS_StringSearch &ss);
	RJS_SubString find(const RJS_StringSearch &ss, int &pos, int &n);
	RJS_SubString find(const RJS_StringSearch &ss, RJS_String &matched); // s=matched string
	RJS_SubString find(const RJS_StringSearch &ss, RJS_String &matched, int &pos, int &n);

	RJS_SubString match(char ch);
	RJS_SubString match(char ch,int &pos);
	RJS_SubString match(const char *s);
	RJS_SubString match(const char *s,int &pos);
	RJS_SubString match(RJS_String &s);
	RJS_SubString match(RJS_String &s,int &pos);
	RJS_SubString match(const RJS_StringSearch &ss);
	RJS_SubString match(const RJS_StringSearch &ss,int &pos, int &n);
	RJS_SubString match(const RJS_StringSearch &ss, RJS_String &matched); // s=matched string
	RJS_SubString match(const RJS_StringSearch &ss, RJS_String &matched, int &pos, int &n);

	int split(const RJS_StringSearch &ss, RJS_String result[], int n);

	RJS_String &trim(Side pt=RJS_String::Both);	// remove white space
	RJS_String &pad(int len,Side pt=RJS_String::Right,char pc=' '); // pads
	RJS_String &trunc(int len);			// truncates length

	RJS_String &upper();			// upper case string
	RJS_String &lower();			// lower case string
	RJS_String &reverse();			// reverse the string

// Access/Misc functions

	virtual void update()		{ if (uf) uf(this);	}
	virtual RJS_String &Str()		{ return *this;		}
	RJS_StringRange length() const 	{ return sd.len;	}
	RJS_StringRange set_length(RJS_StringRange n);
	RJS_StringRange size() const 	{ return sd.size;	}
	operator const char *() const 	{ return sd.data; 	}
        char *cptr() const 		{ return sd.data; 	}
	int empty() const  		{ return length()==0;	}
	operator !() const 		{ return length()==0; 	}
	operator void *() const 	{ return (void *) length(); }
	RJS_String &icase() 		{ RJS_String::temp_icase=1; return *this;}
	RJS_String &ucase() 		{ RJS_String::temp_case=1; return *this; }

// Class Utility Functions
public:

  static void set_update(UpdateFunc f)	{ uf=f;	}
  static RJS_StringRange get_hunksize()	{ return RJS_String::hunksize; }
  static void set_hunksize(RJS_StringRange s){ if (s>0) RJS_String::hunksize=s; }
  static CaseCompare get_case()		{ return RJS_String::use_case;	}
  static void   set_case(CaseCompare c)	{ RJS_String::use_case = c;		}

  void dump() const;			// dump out string contents
  void VERIFY() const;			// verify internals
  void exception(char *message) const;	// abort and dump string

  static void new_error(char *message);	// abort
  static RJS_StringData alloc(const char *data, int len, int size=-1);
  static void resize(RJS_StringData &sd, int new_size);

  static int length(const char *s);
  int search(const char *search_str,int sslen) const;
  int compare(char ch) const;
  int compare(const char *s2) const;
  int compare(const RJS_String &s2) const;
  static int compare(const char *s1,int s1_len,const char *s2,int s2_len);

// Class Private Data/Functions

private:
	virtual void assign(const char *data, RJS_StringRange len);
	static UpdateFunc uf;		// update hook
	static RJS_StringRange hunksize; 	// number of bytes to increase string
					// when it overflows

	static CaseCompare use_case;	// case comparsions (1 use case)
	static char temp_icase;		// temp. ignore case comparsions
	static char temp_case;		// temp. user case comparsions

	RJS_StringData sd;		// this is the actual RJS_String!

};


class RJS_SubString : public RJS_String {
	friend class RJS_String;
public:
	~RJS_SubString();
	RJS_SubString &operator = (char ch);
	RJS_SubString &operator = (const char *s);
	RJS_SubString &operator = (const RJS_String &s);
	RJS_SubString &operator = (const RJS_SubString &s);

	virtual RJS_SubString substr(int p, int l);
	RJS_SubString substr(int p);
	virtual void update() { if (RJS_String::uf) RJS_String::uf(str); }
	virtual RJS_String &Str() { return *str; }
	RJS_SubString &pos(int &pos);

private:
	virtual void assign(const char *ndata,RJS_StringRange nlen);
	RJS_SubString(const RJS_SubString &ss);
        RJS_SubString(RJS_String *s, int p, RJS_StringRange l);
        RJS_SubString(RJS_String *s, int p, RJS_StringRange l, 
		   RJS_String *parent_string, RJS_SubString *parent_substring);
	RJS_String    *str;			// string this is a substring of
	RJS_SubString *sstr;		// substring this is a substring of!
};

typedef int (*RJS_StringSearchFunc)(const RJS_String &s, int &len);

class RJS_StringSearch { 
public:
	RJS_StringSearch() { ssf=NULL; str_flag=0; }
	RJS_StringSearch(const RJS_String &s) { ssf=NULL; str=s; str_flag=1; }
	RJS_StringSearch(RJS_StringSearchFunc func) { ssf=func; str_flag=0;}
	virtual int search(const RJS_String &s, int &matchlen) const ;
private:
	RJS_StringSearchFunc ssf;
	RJS_String str; // used if search through a string
	int str_flag;
};
	
class RJS_StringIterator {
public:
	RJS_StringIterator(const RJS_String &str) :s(str),p(-1) {}
	int operator()(char &ch) {
		if (p < int(s.length()-1)) { p++; ch=s[p]; return 1; }
		else { p=s.length(); return 0; }
	}
	int pos() { return p; }
private:
	const RJS_String &s;
	int p;
};

class RJS_StringScan {
  public:
	enum ScanType { ByField,ByMatch };
	RJS_StringScan(RJS_String &s, ScanType t, const RJS_StringSearch &s_s) 
		: ss(&s_s),st(t),pos(0),lf(0),str(&s) {}
  	int operator()(RJS_String &match);
	int cursor() { return pos; }
  private:
	RJS_String   *str;		// RJS_String we a scanning through
	const RJS_StringSearch *ss;	// pointer to RJS_StringSearch
	ScanType   st;		// type of scan we a preforming
	RJS_StringRange pos;	// current position within string
        char	   lf;		// true if we already returned last field
};

extern const RJS_StringSearch SSwhite;
extern const RJS_StringSearch SSoptwhite;
extern const RJS_StringSearch SSnonwhite;
extern const RJS_StringSearch SSint;

extern const RJS_StringSearch SSalpha;
extern const RJS_StringSearch SSlower;
extern const RJS_StringSearch SSupper;
extern const RJS_StringSearch SSalphanum;
extern const RJS_StringSearch SSid;
extern const RJS_StringSearch SSstr;
extern const RJS_StringSearch SSstrq;
extern const RJS_StringSearch SSunix_word;
extern const RJS_StringSearch SSdcl_word;

extern RJS_String NullRJS_String;		// null RJS_SubStrings point here

// Misc functions that aren't RJS_String members

int    split(RJS_String &s,const RJS_StringSearch &ss, RJS_String result[], int n);

RJS_String operator *(const RJS_String &s, int mult);
RJS_String operator *(int mult,const RJS_String &s);

RJS_String upper(const char *s);	// upper case string
RJS_String upper(const RJS_String &s);

RJS_String lower(const char *s);	// lower case string
RJS_String lower(const RJS_String &s);

RJS_String reverse(const char *s);	// reverse string
RJS_String reverse(const RJS_String &s);

RJS_String trim(const char *s,RJS_String::Side pt=RJS_String::Both);// trim whitespace
RJS_String trim(const RJS_String &s,RJS_String::Side pt=RJS_String::Both);

RJS_String pad(const char *s,int len,RJS_String::Side pt=RJS_String::Right,char pc=' ');
RJS_String pad(const RJS_String &s,int len,RJS_String::Side pt=RJS_String::Right,char pc=' ');

RJS_String trunc(const char *s, int len);
RJS_String trunc(const RJS_String &s, int len);


// inlines

extern char UpperTable[];
extern char LowerTable[];

inline char ToUpper(char c) { return UpperTable[c]; }
inline char ToLower(char c) { return LowerTable[c]; }

inline RJS_String::RJS_String(char *s,RJS_StringRange len,RJS_StringRange size) {
		sd.data=s; sd.len=len; sd.size=size;
}

inline RJS_SubString::~RJS_SubString() { sd.data=0; }
inline RJS_SubString RJS_String::substr(int p)	   { return substr(p,length()-p); }
inline RJS_SubString RJS_SubString::substr(int p ) { return substr(p,length()-p); }
inline RJS_SubString RJS_String::skip() 	   { return skip(SSwhite); }
inline RJS_SubString RJS_String::ws()   	   { return skip(SSwhite); }

// relational stuff

inline int operator <  (char c, const RJS_String &s1)       { return s1 >= c; }
inline int operator <  (const char *c,const RJS_String &s1) { return s1 >= c; }
inline int operator <= (char c, const RJS_String &s1)       { return s1 >= c;  }
inline int operator <= (const char *c,const RJS_String &s1) { return s1 >= c;  }
inline int operator == (char c, const RJS_String &s1)       { return s1 == c; }
inline int operator == (const char *c,const RJS_String &s1) { return s1 == c; }
inline int operator != (char c, const RJS_String &s1)       { return s1 != c; }
inline int operator != (const char *c,const RJS_String &s1) { return s1 != c; }
inline int operator >  (char c, const RJS_String &s1)       { return s1 <= c; }
inline int operator >  (const char *c,const RJS_String &s1) { return s1 <= c; }
inline int operator >= (char c, const RJS_String &s1)       { return s1 <= c;  }
inline int operator >= (const char *c,const RJS_String &s1) { return s1 <= c;  }


inline RJS_SubString RJS_String::left(int n) { 
	return substr((n>0)? 0: -1,n); 
}

inline RJS_SubString RJS_String::right(int n){ 
	return substr((n>0) ? (length()-n) : -1,n); 
}

inline RJS_SubString RJS_String::between(int p1,int p2) {
	return substr((p2>=p1) ? p1 : -1 ,p2-p1+1);
}


inline RJS_String &RJS_String::insert(int pos,char c) { 
	substr(pos,0)=c; return *this;
}

inline RJS_String &RJS_String::insert(int pos,const char *s) {
	substr(pos,0)=s; return *this;
}

inline RJS_String &RJS_String::insert(int pos,const RJS_String &s) {
	substr(pos,0)=s; return *this;
}

inline RJS_String &RJS_String::prepend(char c) {
	substr(0,0)=c; return *this;
}

inline RJS_String &RJS_String::prepend(const char *s) {
	substr(0,0)=s; return *this;
}

inline RJS_String &RJS_String::prepend(const RJS_String &s) {
	substr(0,0)=s; return *this;
}

inline RJS_String &RJS_String::append(char c) {
	substr(length(),0)=c; return *this;
}

inline RJS_String &RJS_String::append(const char *s) {
	substr(length(),0)=s; return *this;
}

inline RJS_String &RJS_String::append(const RJS_String &s) {
	substr(length(),0)=s; return *this;
}

inline RJS_String &RJS_String::remove(int pos, int n) {
	substr(pos,n)=""; return *this;
}

inline RJS_String &RJS_String::remove(char c) {
	at(c)=""; return *this;
}	

inline RJS_String &RJS_String::remove(const char *s) {
	at(s)=""; return *this;
}

inline RJS_String &RJS_String::remove(const RJS_String &s) {
	at(s)=""; return *this;
}

inline RJS_String &RJS_String::remove(const RJS_StringSearch &ss) {
	at(ss)=""; return *this;
}

inline int RJS_String::index(char ch) const {  
	return search(&ch,1); 
}

inline int RJS_String::index(const char *s) const { 
	return search(s,RJS_String::length(s)); 
}

inline int RJS_String::index(const RJS_String &s) const {   
		return search(s.cptr(),s.length()); 
}

inline int RJS_String::index(const RJS_StringSearch &ss) const {
		int ss_len;
		return ss.search(*this,ss_len);
}

inline int RJS_String::contains(char ch) const { 
	return search(&ch,1)!= -1; 
}

inline int RJS_String::contains(const char *s) const { 
	return search(s,RJS_String::length(s)) != -1;
}

inline int RJS_String::contains(const RJS_String &s) const {	
	return search(s.cptr(),s.length()) != -1; 
}

inline int RJS_String::contains(const RJS_StringSearch &ss) const { 
	int ss_len;
	return ss.search(*this,ss_len)!= -1; 
}

inline RJS_String &RJS_String::operator+=(char c) {
	substr(length(),0)=c; return *this;
}

inline RJS_String &RJS_String::operator+=(const char *s) {
	substr(length(),0)=s; return *this;
}

inline RJS_String &RJS_String::operator+=(const RJS_String &s) {
	substr(length(),0)=s; return *this;
}


#endif
